from django.apps import AppConfig
from django.utils.translation import pgettext_lazy


__all__ = ('LtreeConfig',)


class LtreeConfig(AppConfig):
    name = 'pxd_postgres.ltree'
    label = 'pxd_postgres_ltree'
    verbose_name = pgettext_lazy('pxd_postgres:ltree', 'Postgresql ltree utils')



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy


__all__ = ('PostgresConfig',)


class PostgresConfig(AppConfig):
    name = 'pxd_postgres'
    verbose_name = pgettext_lazy('pxd_postgres', 'Postgresql utils')



from django.forms import CharField

from .validators import ltree_validator, ltree_int_validator


__all__ = 'LtreeFormField', 'LtreeIntFormField',


class LtreeFormField(CharField):
    default_validators = [ltree_validator]


class LtreeIntFormField(CharField):
    default_validators = [ltree_int_validator]



from django.db import migrations


CREATE_EXTENSION_SQL = 'CREATE EXTENSION IF NOT EXISTS ltree;'
DROP_EXTENSION_SQL = 'DROP EXTENSION IF EXISTS ltree;'

# Function to transform ltree to an array of strings.
# Delimiter - is the default delimiter `.` as in ltree.
CREATE_LTREE_TO_ARRAY_FUNCTION_SQL = [
    '''--sql
    -- Splitting ltree by path separator(.)
    CREATE OR REPLACE FUNCTION ltree2array(item ltree) RETURNS text[] AS
    $BODY$
        SELECT string_to_array(ltree2text(item), '.')
    $BODY$
    LANGUAGE SQL;
    ''',

    '''--sql
    -- Splitting text as ltree
    CREATE OR REPLACE FUNCTION ltree2array(item text) RETURNS text[] AS
    $BODY$
        SELECT string_to_array(item, '.')
    $BODY$
    LANGUAGE SQL;
    '''
]
DROP_LTREE_TO_ARRAY_FUNCTION_SQL = 'DROP FUNCTION IF EXISTS ltree2array;'

# Function to transform array of strings to ltree.
# Delimiter - is the default delimiter `.` as in ltree.
CREATE_ARRAY_TO_LTREE_FUNCTION_SQL = '''--sql
-- Splitting ltree by path separator(.)
CREATE OR REPLACE FUNCTION array2ltree(item text[]) RETURNS ltree AS
$BODY$
    SELECT text2ltree(array_to_string(item, '.'))
$BODY$
LANGUAGE SQL;
'''

DROP_ARRAY_TO_LTREE_FUNCTION_SQL = 'DROP FUNCTION IF EXISTS array2ltree;'

# Function to `unnest(item ltree)` into a containing elements.
CREATE_LTREE_UNNEST_FUNCTION_SQL = '''--sql
-- Splitting ltree by path separator(.)
CREATE OR REPLACE FUNCTION unnest(item ltree) RETURNS SETOF text AS
$BODY$
BEGIN
    RETURN QUERY SELECT unnest(ltree2array(item));
    RETURN;
END;
$BODY$
LANGUAGE plpgsql;
'''
DROP_LTREE_UNNEST_FUNCTION_SQL = 'DROP FUNCTION IF EXISTS unnest(ltree);'


class Migration(migrations.Migration):
    dependencies = []
    operations = [
        migrations.RunSQL(CREATE_EXTENSION_SQL, DROP_EXTENSION_SQL),
        migrations.RunSQL(
            CREATE_LTREE_TO_ARRAY_FUNCTION_SQL,
            DROP_LTREE_TO_ARRAY_FUNCTION_SQL,
        ),
        migrations.RunSQL(
            CREATE_ARRAY_TO_LTREE_FUNCTION_SQL,
            DROP_ARRAY_TO_LTREE_FUNCTION_SQL,
        ),
        migrations.RunSQL(
            CREATE_LTREE_UNNEST_FUNCTION_SQL,
            DROP_LTREE_UNNEST_FUNCTION_SQL,
        ),
    ]



from typing import Type
from django.db.models.fields import TextField
from django.forms.widgets import TextInput

from .value import LtreeValueParserables, LtreeValue, LtreeIntValue
from .validators import ltree_validator, ltree_int_validator
from .forms import LtreeFormField, LtreeIntFormField


__all__ = (
    'LtreeValueProxy',
    'LtreeField',
    'LtreeIntValueProxy',
    'LtreeIntField',
)


class LtreeValueProxy:
    _value_type: Type[LtreeValue] = LtreeValue

    def __init__(self, field_name: str) -> None:
        self.field_name = field_name

    def __get__(self, instance, owner):
        if instance is None:
            return self

        value = instance.__dict__[self.field_name]

        if value is None:
            return value

        return self._value_type(instance.__dict__[self.field_name])

    def __set__(self, instance, value):
        if instance is None:
            return self

        instance.__dict__[self.field_name] = value


class LtreeField(TextField):
    default_validators = [ltree_validator]
    _value_type: Type[LtreeValue] = LtreeValue
    _value_proxy_type: Type[LtreeValueProxy] = LtreeValueProxy
    _form_class = LtreeFormField
    _widget = TextInput

    def db_type(self, connection):
        return 'ltree'

    def formfield(self, **kwargs):
        kwargs['form_class'] = self._form_class
        kwargs['widget'] = self._widget()
        return super(LtreeField, self).formfield(**kwargs)

    def contribute_to_class(self, cls, name, private_only=False):
        super(LtreeField, self).contribute_to_class(cls, name)
        setattr(cls, self.name, self._value_proxy_type(self.name))

    def from_db_value(self, value, expression, connection, *args):
        if value is None:
            return value
        return self._value_type(value)

    def get_prep_value(self, value):
        if value is None:
            return value
        return str(self._value_type(value))

    def to_python(self, value):
        if value is None:
            return value
        elif isinstance(value, self._value_type):
            return value

        return self._value_type(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        if value is None:
            return value
        elif isinstance(value, self._value_type):
            return str(value)
        elif isinstance(value, LtreeValueParserables):
            return str(self._value_type(value))

        raise ValueError('Unknown value type {}'.format(type(value)))


class LtreeIntValueProxy(LtreeValueProxy):
    _value_type: Type[LtreeIntValue] = LtreeIntValue


class LtreeIntField(LtreeField):
    default_validators = [ltree_int_validator]
    _value_type: Type[LtreeIntValue] = LtreeIntValue
    _value_proxy_type: Type[LtreeIntValueProxy] = LtreeIntValueProxy
    _form_class = LtreeIntFormField



from collections.abc import Iterable
from collections import UserList
from typing import Union


__all__ ='LtreeValueParserables', 'LtreeValue', 'LtreeIntValue',


LtreeValueParserables = (str, int, float, Iterable)


class LtreeValue(UserList):
    _base_type: type = str

    def __init__(self, value: Union[Iterable, str, int]) -> None:
        base = self._base_type

        if isinstance(value, str):
            value = value.strip().split('.') if value else []
        elif isinstance(value, (int, float)):
            value = [int(value)]
        elif isinstance(value, Iterable):
            value = [v for v in value]
        else:
            raise ValueError('Invalid value: {!r} for ltree'.format(value))

        super(LtreeValue, self).__init__(initlist=[base(x) for x in value])

    def __repr__(self):
        return str(self)

    def __str__(self):
        return '.'.join(map(str, self))


class LtreeIntValue(LtreeValue):
    _base_type: type = int



from django.db.models import Func

from .fields import LtreeField

__all__ = 'LtreeConcatFunc', 'LtreeSubpathFunc',


class LtreeConcatFunc(Func):
    function = 'ltree_concat'
    arg_joiner = ' || '
    template_postgresql = "%(expressions)s"

    def as_postgresql(self, compiler, connection):
        return self.as_sql(compiler, connection, template=self.template_postgresql)


class LtreeSubpathFunc(Func):
    function = 'subpath'

    @property
    def output_field(self):
        return LtreeField()



from django.core.validators import RegexValidator
from django.utils.translation import pgettext_lazy


__all__ = 'ltree_validator', 'ltree_int_validator'


ltree_validator = RegexValidator(
    r'^[A-Za-z0-9_.]*$',
    pgettext_lazy(
        'pxd_postgres',
        'Postgresql ltree is a sequence of alphanumeric characters and '
        'underscores separated by dots.',
    ),
    'invalid',
)

ltree_int_validator = RegexValidator(
    r'^[0-9.]*$',
    pgettext_lazy(
        'pxd_postgres',
        'Int ltree is a sequence of numbers separated by dots.',
    ),
    'invalid',
)



from django.db.models import Lookup, Transform, IntegerField

from .fields import LtreeField


__all__ = (
    'EqualLookup',
    'AncestorLookup',
    'DescendantLookup',
    'MatchLookup',
    'ContainsLookup',

    'DepthTransform',
)


class SimpleLookup(Lookup):
    lookup_operator: str = '='

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = list(lhs_params) + list(rhs_params)
        return '{} {} {}'.format(lhs, self.lookup_operator, rhs), params


@LtreeField.register_lookup
class EqualLookup(SimpleLookup):
    lookup_name: str = 'exact'
    lookup_operator: str = '='


@LtreeField.register_lookup
class AncestorLookup(SimpleLookup):
    lookup_name: str = 'ancestor'
    lookup_operator: str = '@>'


@LtreeField.register_lookup
class DescendantLookup(SimpleLookup):
    lookup_name: str = 'descendant'
    lookup_operator: str = '<@'


@LtreeField.register_lookup
class MatchLookup(SimpleLookup):
    lookup_name: str = 'match'
    lookup_operator: str = '~'


@LtreeField.register_lookup
class ContainsLookup(SimpleLookup):
    lookup_name: str = 'contains'
    lookup_operator: str = '?'


@LtreeField.register_lookup
class DepthTransform(Transform):
    lookup_name = 'depth'
    function = 'nlevel'

    @property
    def output_field(self):
        return IntegerField()
